#!/usr/bin/env perl

use strict;
use warnings;

use FindBin;
use File::Basename;
use Cwd;
use Carp;
use Getopt::Long qw(:config no_ignore_case bundling);


my $usage = <<_EOUSAGE_;

##################################################################
#
# Required:
#
#  --seqType <string>  :type of reads: (fq or fa)
#
#  If paired reads:
#
#  --left  <string>    :left reads
#  --right <string>    :right reads
# 
#  If unpaired reads:
#
#  --single <string>   :single reads
#
#
#  if strand-specific data, set:
#  
#  --SS_lib_type <string>  :if paired: RF or FR,  if single: F or R
#
#  
#  --output <string>     :name of directory for output (will be created if it doesn't already exist)
#
#
#  Butterfly options:
# 
#  --run_butterfly                 :executes butterfly commands.  Do not set this if you want to spawn them on a computing grid.   
#        
#  (if run_butterfly) 
#       --num_butterfly_CPU <int>  :number of parallel butterfly commands to run on a single machine (default: 1)  
#
# Misc:
#
# 
#  --min_contig_length <int> :minimum assembled contig length to report (def=300)
#
#  --paired_fragment_length <int>  :size of a read pair insert (def=300)
#
#  --jaccard_clip     :option, set if you have paired reads and you expect high gene density with UTR overlap (use FASTQ input files for reads).
#
##############################################################################################


_EOUSAGE_

	;



my $advanced_options = <<__ADVANCED__;

#
#  --inchworm <string>   :start with existing inchworm assemblies (skips the inchworm step given existing output)  
#  --USE_READS <string>  :use the fasta reads already extracted
#

__ADVANCED__
	
	;

my $INCHWORM_DIR = "$FindBin::Bin/Inchworm";
my $CHRYSALIS_DIR = "$FindBin::Bin/Chrysalis";
my $BUTTERFLY_DIR = "$FindBin::Bin/Butterfly";

# option list:
my ($seqType, $left_file, $right_file, $single_file, $SS_lib_type, $min_contig_length,
	$paired_fragment_length, $jaccard_clip, $inchworm_file, $show_advanced_options,
	$output_directory,
	);

# defaults:
$min_contig_length = 300;
$paired_fragment_length = 300;


my $RUN_BUTTERFLY_FLAG = 0;
my $NUM_BUTTERFLY_CPU = 1;

## ADVANCED OPTIONS:
     # todo  add some.

my $use_reads;





&GetOptions( 
			 "seqType=s" => \$seqType,
			 "left=s" => \$left_file,
			 "right=s" => \$right_file,
			 "single=s" => \$single_file,

			 "SS_lib_type=s" => \$SS_lib_type,
			
			 "output=s" => \$output_directory,

			 "min_contig_length=i" => \$min_contig_length,
			 "paired_fragment_length=i" => \$paired_fragment_length,
			 "jaccard_clip" => \$jaccard_clip,
			 
			 
			 'run_butterfly' => \$RUN_BUTTERFLY_FLAG,
			 'num_butterfly_CPU=i' => \$NUM_BUTTERFLY_CPU,

			 "show_advanced_options" => \$show_advanced_options,
			 
			 

			 ## Advanced options
			 
			 "inchworm=s" => \$inchworm_file,
			 "USE_READS=s" => \$use_reads,
			 

			 );


## Check options set:

unless ($seqType && 
		( 
		  ($left_file && $right_file) || $single_file || $use_reads) 
		) {
	die $usage;
}



unless ($output_directory) {
	$output_directory = "trinity_dir.$$";
}

main: {

	my $start_dir = cwd();

	## create complete paths for input files:
	$left_file = &create_full_path($left_file) if $left_file;
	$right_file = &create_full_path($right_file) if $right_file;
	$single_file = &create_full_path($single_file) if $single_file;
	$inchworm_file = &create_full_path($inchworm_file) if $inchworm_file;
	$output_directory = &create_full_path($output_directory);
	
	$use_reads = &create_full_path($use_reads) if $use_reads;

	unless (-d $output_directory) {
		
		mkdir $output_directory or die "Error, cannot mkdir $output_directory";
	}
	
	chdir ($output_directory) or die "Error, cannot cd to $output_directory";
	
	my $trinity_target_fa = "";
	
	if ($use_reads) {
		$trinity_target_fa = "reads.fa";
		&process_cmd("ln -s $use_reads $trinity_target_fa");
	}
	elsif ($left_file && $right_file) {
		my ($left_SS_type, $right_SS_type);
		if ($SS_lib_type) {
			($left_SS_type, $right_SS_type) = split(//, $SS_lib_type);
		}
		
		&prep_seqs($left_file, $seqType, "left", $left_SS_type);
		&prep_seqs($right_file, $seqType, "right", $right_SS_type);
		
		$trinity_target_fa = "both.fa";
		&process_cmd("cat left.fa right.fa > $trinity_target_fa");
		
	}
	elsif ($single_file) {
		&prep_seqs($single_file, $seqType, "single", $SS_lib_type);
		$trinity_target_fa = "single.fa";
	}
	
	else {
		die "not sure what to do. "; # should never get here.
	}
	
	#################
	## Inchworm step:
	
	unless ($inchworm_file) {
		$inchworm_file = &run_inchworm($trinity_target_fa, $SS_lib_type);
		
	}
	
	if ($jaccard_clip && $left_file && $right_file) {
		$inchworm_file = &run_jaccard_clip($inchworm_file, $left_file, $right_file, $SS_lib_type);
	}
	
	
	
	##################
	## Chrysalis step:
	
	my $butterfly_cmds = &run_chrysalis($inchworm_file, $trinity_target_fa,
										$min_contig_length, $paired_fragment_length, $SS_lib_type);



	print "Inchworm and Chrysalis complete.  Butterfly commands to execute are provided here:\n"
		. $butterfly_cmds . "\n\n";
	


	if ($RUN_BUTTERFLY_FLAG) {

		my $cmd = "$FindBin::Bin/util/cmd_process_forker.pl -c $butterfly_cmds --CPU $NUM_BUTTERFLY_CPU";
		&process_cmd($cmd);

		
		## capture results:
		$cmd = 'find ./chrysalis -regex ".*allProbPaths.fasta" -exec cat {} \; > Trinity.fasta';

		&process_cmd($cmd);

		print "\n\nButterfly assemblies are written to Trinity.fasta\n\n";
		

	}
	
	exit(0);
	
}


####
sub run_chrysalis {
	my ($inchworm_file, $reads_file,
		$min_contig_length, $paired_fragment_length, $SS_lib_type) = @_;

	my $cmd = "$CHRYSALIS_DIR/Chrysalis -i $reads_file -iworm $inchworm_file -o chrysalis "
		. " -min $min_contig_length -dist $paired_fragment_length ";
	
	if ($SS_lib_type) {
		$cmd .= " -strand 1 ";
	}
	
	$cmd .= " -butterfly $BUTTERFLY_DIR/Butterfly.jar ";
	
	&process_cmd($cmd);

	my $butterfly_cmds = &create_full_path("chrysalis/butterfly_commands");
	unless (-s $butterfly_cmds) {
		croak "Error, chrysalis did not report butterfly commands file: $butterfly_cmds";
	}
	
	return($butterfly_cmds);
	
}


####
sub run_inchworm {
	my ($reads, $strand_specific_flag) = @_;
	
	my $inchworm_outfile = "inchworm.K25.L48";

	my $inchworm_cmd = "$INCHWORM_DIR/bin/inchworm --reads $reads --run_inchworm -K 25 -L 48 --monitor 1 ";
    unless ($strand_specific_flag) {
		$inchworm_cmd .= " --DS ";
		$inchworm_outfile .= ".DS";
	}
	
	$inchworm_outfile .= ".fa";
	
	$inchworm_cmd .= " 2>monitor.out > $inchworm_outfile";
	
	&process_cmd($inchworm_cmd) unless (-s $inchworm_outfile);

	return($inchworm_outfile);
	
}

####
sub prep_seqs {
	my ($initial_file, $seqType, $file_prefix, $SS_lib_type) = @_;

	if ($seqType eq "fq") {
		# make fasta
		
		my $cmd = "$INCHWORM_DIR/util/fastQ_to_fastA.pl -I $initial_file ";
		if ($SS_lib_type && $SS_lib_type eq "R") {
			$cmd .= " --rev ";
		}
		$cmd .= "> $file_prefix.fa";
		
		&process_cmd($cmd) unless (-e "$file_prefix.fa");
	}
	elsif ($seqType eq "fa") {
		if ($SS_lib_type && $SS_lib_type eq "R") {
			my $cmd = "$INCHWORM_DIR/util/revcomp_fasta.pl $initial_file > $file_prefix.fa";
			&process_cmd($cmd) unless (-s "$file_prefix.fa");
		}
		else {
			## just symlink it here:
			my $cmd = "ln -s $initial_file $file_prefix.fa";
			&process_cmd($cmd) unless (-s "$file_prefix.fa");
		}
	}
	
	return;
}



###
sub create_full_path {
	my ($file) = @_;

	my $cwd = cwd();
	if ($file !~ m|^/|) { # must be a relative path
		$file = $cwd . "/$file";
	}
	
	return($file);
}



####
sub process_cmd {
	my ($cmd) = @_;

	print "CMD: $cmd\n";

	my $ret = system($cmd);

	if ($ret) {
		die "Error, cmd: $cmd died with ret $ret";
	}
	
	return;
}


####
sub run_jaccard_clip {
	my ($inchworm_file, $left_file, $right_file, $SS_lib_type) = @_;

	my $cmd = "$INCHWORM_DIR/bin/inchworm_transcript_splitter.pl --iworm $inchworm_file "
		. " --left_fq $left_file --right_fq $right_file ";

	if ($SS_lib_type) {
		$cmd .= " --SS_lib_type $SS_lib_type ";
	}
	
	&process_cmd($cmd);

	my $output_file = "$inchworm_file.clipped.fa";

	unless (-s $output_file) {
		croak "Error, jaccard clipping didn't produce the expected output file: $output_file";
	}

	return($output_file);
}
